// engine.cpp
/*
 *  Copyright (c) 2008-2011 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "stdafx.h"
#include <fstream>
#include "engine.h"

engine::engine() :
	iCurrentFile(iFiles.end()),
	iCurrentFileSize(0), iCurrentFileRead(0),
	iFileCount(0),
	iFailures(0),
	iCharacterCount(0),
	iLineCount(0)
{
}

engine::~engine()
{
	abort();
}

bool engine::any_left()
{
	lock();
	bool ret = false;
	if (is_running())
	{
		ret = (file_list::const_iterator(iCurrentFile) != iFiles.end());
		if (!ret)
			iFiles.erase(iFiles.begin(), iCurrentFile);
	}
	unlock();
	return ret;
}

bool engine::analyze_file(const std::string& aFile, unsigned long& aLineCount, unsigned long& aCharacterCount)
{
	iCurrentFileSize = 0;
	iCurrentFileRead = 0;

	std::ifstream in(aFile.c_str(), std::ios::in|std::ios::binary);
	if (!in)
		return false;

	in.seekg(0, std::ios::end);
	iCurrentFileSize = static_cast<unsigned long>(in.tellg());
	in.seekg(0, std::ios::beg);

	aLineCount = 0;
	aCharacterCount = 0;

	char character = 0;
	while(in && is_running())
	{
		char buffer[1024];
		in.read(buffer, sizeof(buffer));
		std::size_t count = in.gcount();
		if (count > 0)
		{
			for (std::size_t i = 0; i < count; ++i)
			{
				++aCharacterCount;
				++iCurrentFileRead;
				character = buffer[i];
				if (character == '\n')
					++aLineCount;
			}
		}
	}
	if (aCharacterCount > 0 && character != '\n')
		++aLineCount;

	return true;
}

void engine::analyze_files()
{
	while(any_left())
	{
		unsigned long lines, characters;
		bool ok = analyze_file(*iCurrentFile, lines, characters);
		lock();
		++iCurrentFile;
		if (ok)
		{
			iLineCount += lines;
			iCharacterCount += characters;
			++iFileCount;
		}
		else
			++iFailures;
		unlock();
	}
}

void engine::add_file(const std::string& aFile)
{
	lock();
	iFiles.push_back(aFile);
	if (iCurrentFile == iFiles.end())
		--iCurrentFile;
	if (!is_running())
		run();
	unlock();
}

unsigned int engine::progress() const
{
	lock();
	file_list::size_type current = std::distance(iFiles.begin(), file_list::const_iterator(iCurrentFile));
	unsigned int percentage = iFiles.size() != 0 ? (current * 100) / iFiles.size() : 0;
	unlock();
	return percentage;
}

unsigned int engine::file_progress() const
{
	lock();
	unsigned int percentage = static_cast<unsigned int>(iCurrentFileSize != 0 ? (iCurrentFileRead * 100.0) / iCurrentFileSize : 0);
	unlock();
	return percentage;
}

void engine::results(unsigned long& aFileCount, unsigned long& aFailures, unsigned long& aCharacterCount, unsigned long& aLineCount) const
{
	lock();
	aFileCount = iFileCount;
	aFailures = iFailures;
	aCharacterCount = iCharacterCount;
	aLineCount = iLineCount;
	unlock();
}

void engine::reset()
{
	abort();
	iCurrentFileSize = 0;
	iCurrentFileRead = 0;
	iCharacterCount = 0;
	iLineCount = 0;
	iFiles.clear();
	iCurrentFile = iFiles.begin();
	iFileCount = 0;
	iFailures = 0;
}

bool engine::analyzing() const
{
	lock();
	bool ret = is_running();
	unlock();
	return ret;
}
