// file.cpp - v1.3.1
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "stdafx.h"
#include <windows.h>
#include <WINNLS.H>
#include <shellapi.h>
#include <vector>
#include <string>
#include <cassert>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include "string.h"
#include "file.h"

namespace lib
{
	void tidy_path(std::string& aPath)
	{
		for (std::string::size_type i = 0; i != aPath.size(); ++i)
			if (aPath[i] == '/')
				aPath[i] = '\\';
	}

	void tidy_path(std::wstring& aPath)
	{
		for (std::wstring::size_type i = 0; i != aPath.size(); ++i)
			if (aPath[i] == L'/')
				aPath[i] = L'\\';
	}

	std::string make_path(const std::wstring& aString)
	{
		std::vector<char> name;
		name.resize(MAX_PATH+1);
		BOOL useDefault;
		WideCharToMultiByte(CP_ACP, WC_NO_BEST_FIT_CHARS, aString.c_str(), aString.length(), &name[0], name.size(), NULL, &useDefault);
		if (useDefault) 
		{
			std::vector<wchar_t> shortname;
			shortname.resize(MAX_PATH+1);
			GetShortPathNameW(aString.c_str(), &shortname[0], shortname.size());
			WideCharToMultiByte(CP_ACP, WC_NO_BEST_FIT_CHARS, &shortname[0], shortname.size(), &name[0], name.size(), NULL, NULL);
		}
		return std::string(&name[0]);
	}

	std::wstring make_path(const std::string& aString)
	{
		return lib::utf8_to_wide(aString);
	}

	void create_path(const std::string& aPath)
	{
		std::vector<std::string> parts;
		lib::tokens(aPath, std::string("\\"), parts);
		std::string partPath;
		for (std::vector<std::string>::const_iterator i = parts.begin(); i != parts.end(); ++i)
		{
			partPath += *i;
			partPath += "\\";
			CreateDirectoryW(make_path(partPath).c_str(), NULL);
		}
	}

	void create_path(const std::wstring& aPath)
	{
		std::vector<std::wstring> parts;
		lib::tokens(aPath, std::wstring(L"\\"), parts);
		std::wstring partPath;
		for (std::vector<std::wstring>::const_iterator i = parts.begin(); i != parts.end(); ++i)
		{
			partPath += *i;
			partPath += L"\\";
			CreateDirectoryW(partPath.c_str(), NULL);
		}
	}

	std::string create_file(const std::string& aFileName)
	{
		std::wstring wideFileName = lib::utf8_to_wide(aFileName);
		lib::create_file(wideFileName);
		return lib::make_path(wideFileName);
	}

	void create_file(const std::wstring& aFileName)
	{
		HANDLE hFile = CreateFile(aFileName.c_str(), FILE_ALL_ACCESS, 0, NULL, CREATE_NEW, FILE_ATTRIBUTE_NORMAL, NULL);
		if (hFile != NULL)
			CloseHandle(hFile);
	}

	bool file_exists(const std::string& aPath)
	{
		return file_exists(make_path(aPath));
	}

	bool can_read_file(const std::string& aPath)
	{
		return can_read_file(make_path(aPath));
	}

	unsigned long file_size(const std::string& aPath)
	{
		return file_size(make_path(aPath));
	}

	unsigned long long large_file_size(const std::string& aPath)
	{
		return large_file_size(make_path(aPath));
	}

	bool file_exists(const std::wstring& aPath)
	{
		struct _stat fileinfo;
		return _wstat(aPath.c_str(), &fileinfo) == 0;
	}

	bool can_read_file(const std::wstring& aPath)
	{
		FILE* fp = _wfopen(aPath.c_str(), L"rb");
		if (fp != NULL)
		{
			fclose(fp);
			return true;
		}
		return false;
	}

	unsigned long file_size(const std::wstring& aPath)
	{
		struct _stat fileinfo;
		return _wstat(aPath.c_str(), &fileinfo) == 0 ? fileinfo.st_size : 0;
	}

	unsigned long long large_file_size(const std::wstring& aPath)
	{
		struct _stat64 fileinfo;
		return _wstat64(aPath.c_str(), &fileinfo) == 0 ? fileinfo.st_size : 0;
	}

	int large_file_seek(FILE* aStream, long long aOffset, int aOrigin)
	{
		return _fseeki64(aStream, aOffset, aOrigin);
	}

	bool move_file(const std::string& aPathFrom, const std::string& aPathTo)
	{
		return MoveFileA(aPathFrom.c_str(), aPathTo.c_str()) == TRUE;
	}

	void remove_directory(const std::string& aPath)
	{
		SHFILEOPSTRUCT sfo = {};
		sfo.hwnd = NULL;
		sfo.wFunc = FO_DELETE;
		std::wstring path = make_path(aPath);
		path.push_back('\0');
		sfo.pFrom = path.c_str();
		sfo.pTo = NULL;
		sfo.fFlags = FOF_NOCONFIRMATION|FOF_SILENT;
		SHFileOperation(&sfo);
	}

	std::wstring program_directory()
	{
		wchar_t szFullPath[MAX_PATH];
		GetModuleFileName (NULL, szFullPath, sizeof(szFullPath));
		std::wstring programDirectory = szFullPath;
		std::wstring::size_type pos = programDirectory.find_last_of(L'\\');
		assert(pos != std::wstring::npos);
		programDirectory.erase(pos+1);
		return programDirectory;
	}
}