// handle.cpp - v1.3.2
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "stdafx.h"
#include <vector>
#include "handle.h"

namespace lib
{
	const handle::os_handle handle::HandleNone = NULL;
	const handle::os_handle handle::HandleInvalid = INVALID_HANDLE_VALUE;
	const handle::timeout_interval handle::TimeoutInfinite = INFINITE;
	const handle::timeout_interval handle::TimeoutShort = 20;

	handle::handle() : iHandle(HandleNone)
	{
	}
	handle::handle(os_handle aHandle) : iHandle(aHandle)
	{
	}
	handle::handle(const handle& aOther) : iHandle(HandleNone)
	{
		duplicate(aOther);
	}
	handle& handle::operator=(const handle& aOther)
	{
		duplicate(aOther);
		return *this;
	}
	handle& handle::operator=(os_handle aHandle)
	{
		close();
		iHandle = aHandle;
		return*this;
	}
	handle::~handle()
	{
		close();
	}
	const handle* handle::wait(timeout_interval aTimeoutInterval) const
	{
		if (!valid())
			throw std::logic_error("handle::wait");
		timeout_interval result = ::WaitForSingleObject(iHandle, aTimeoutInterval);
		return (result == WAIT_OBJECT_0 ? this : 0);
	}
	const handle* handle::msg_wait(timeout_interval aTimeoutInterval) const
	{
		if (!valid())
			throw std::logic_error("handle::msg_wait");
		timeout_interval result = ::MsgWaitForMultipleObjects(1, &iHandle, FALSE, aTimeoutInterval, QS_ALLEVENTS);
		return (result == WAIT_OBJECT_0 ? this : 0);
	}
	const handle* handle::wait_for_multiple(const handle& aHandle, timeout_interval aTimeoutInterval) const
	{
		handle_ref theRef(aHandle);
		return wait_for_multiple(&theRef, 1, aTimeoutInterval);
	}
	const handle* handle::msg_wait_for_multiple(const handle& aHandle, timeout_interval aTimeoutInterval) const
	{
		handle_ref theRef(aHandle);
		return msg_wait_for_multiple(&theRef, 1, aTimeoutInterval);
	}
	const handle* handle::wait_for_multiple(const handle_ref* aHandleRefs, std::size_t aCount, timeout_interval aTimeoutInterval) const
	{
		std::vector<os_handle> handles;
		handles.reserve(aCount + 1);
		if (aHandleRefs != 0 && aCount != 0)
			handles.insert(handles.end(), aHandleRefs, aHandleRefs + aCount);
		if (valid())
			handles.push_back(iHandle);
		if (handles.empty())
			return 0;
		timeout_interval result = ::WaitForMultipleObjects(handles.size(), &handles[0], FALSE, aTimeoutInterval);
		if (result >= WAIT_OBJECT_0 && result < WAIT_OBJECT_0 + handles.size())
		{
			std::size_t index = result - WAIT_OBJECT_0;
			if (index < aCount)
				return &*aHandleRefs[index];
			else
				return this;
		}
		return 0;
	}
	const handle* handle::msg_wait_for_multiple(const handle_ref* aHandleRefs, std::size_t aCount, timeout_interval aTimeoutInterval) const
	{
		std::vector<os_handle> handles;
		handles.reserve(aCount + 1);
		if (aHandleRefs != 0 && aCount != 0)
			handles.insert(handles.end(), aHandleRefs, aHandleRefs + aCount);
		if (valid())
			handles.push_back(iHandle);
		timeout_interval result = ::MsgWaitForMultipleObjects(handles.size(), !handles.empty() ? &handles[0] : NULL, FALSE, aTimeoutInterval, QS_ALLEVENTS);
		if (result >= WAIT_OBJECT_0 && result < WAIT_OBJECT_0 + handles.size())
		{
			std::size_t index = result - WAIT_OBJECT_0;
			if (index < aCount)
				return &*aHandleRefs[index];
			else
				return this;
		}
		return 0;
	}
	void handle::close()
	{
		if (valid())
		{
			::CloseHandle(iHandle);
			iHandle = HandleNone;
		}
	}
	void handle::duplicate(const handle& aOther)
	{
		close();
		if (aOther.valid())
			::DuplicateHandle(::GetCurrentProcess(), aOther.iHandle, ::GetCurrentProcess(), &iHandle, 0, TRUE, DUPLICATE_SAME_ACCESS);
	}
} // namespace lib