// http.h
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef LIB_HTTP
#define LIB_HTTP

#include <vector>
#include <string>
#include "observable.h"
#include "string.h"
#include "optional.h"
#include "network.h"

namespace lib
{
	class http_observer
	{
		friend class http;
	private:
		virtual void http_request_started(http& aRequest) = 0;
		virtual void http_request_completed(http& aRequest) = 0;
		virtual void http_request_failure(http& aRequest) = 0;
	public:
		enum notify_type { NotifyStarted, NotifyCompleted, NotifyFailure };
	};

	class http : public observable<http_observer>, private socket_manager_observer
	{
		// types
	public:
		typedef std::map<lib::ci_string, std::string> headers_t;
		typedef std::vector<char> body_t;
		enum type_e { Get, Post };

		// construction
	public:
		http(socket_manager& aSocketManager);
		http(const http& aOther);
		virtual ~http();
		http& operator=(const http& aOther);

		// operations
	public:
		void request(const std::string& aUrl, type_e aType = Get, const headers_t& aRequestHeaders = headers_t(), const lib::variant<body_t, std::string>& aRequestBody = std::string());
		void request(const std::string& aHost, const std::string& aResource, type_e aType = Get, u_short aPort = 80, bool aSecure = false, const headers_t& aRequestHeaders = headers_t(), const lib::variant<body_t, std::string>& aRequestBody = std::string());
		bool ok() const { return iOk; }
		unsigned int status_code() const { return iStatusCode; }
		unsigned long body_length() const { return iBodyLength ? *iBodyLength : iBody.size(); }
		const std::string& response_status() const { return iResponseStatus; }
		const headers_t& response_headers() const { return iResponseHeaders; }
		const body_t& body() const { return iBody; }
		std::string body_as_string() const { return std::string(iBody.begin(), iBody.end()); }
		double percent_done() const;

		// implementation
	private:
		void init();
		void add_response_header(const std::string& aHeaderLine);
		bool decode();
		bool decode_chunked();
		// from observable<http_observer>
		virtual void notify_observer(http_observer& aObserver, http_observer::notify_type aType, const void* aParameter, const void* aParameter2);
		// from socket_manager::observer
		virtual void socket_added(socket& aSocket, bool aListener) {}
		virtual void socket_removed(socket& aSocket);
		virtual void socket_connected(socket& aSocket);
		virtual void socket_data(socket& aSocket, const packet& aPacket);
		virtual void socket_underflow(socket& aSocket) {}
		virtual void host_resolved(const std::string& aHostName, const std::string& aHostNumber, u_long aAddress) {}

		// attributes
	private:
		socket_manager& iSocketManager;
		std::string iHost;
		u_short iPort;
		bool iSecure;
		type_e iType;
		std::string iResource;
		headers_t iRequestHeaders;
		body_t iRequestBody;
		socket* iSocket;
		std::string iResponseLine;
		std::string iResponseStatus;
		headers_t iResponseHeaders;
		headers_t::iterator iLastResponseHeader;
		bool iOk;
		unsigned int iStatusCode;
		optional<unsigned long> iBodyLength;
		body_t iBody;
		enum state { ResponseStatus, ResponseHeaders, Body, Finished } iState;
		bool iPreviousWasCRLF;
	};
}

#endif //LIB_HTTP
