// logger.h
/*
 *  Copyright (c) 2011 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* WARNING: The algorithms present here are not a substitute for any equivalent std:: 
 * algorithms available on your platform which you should be using instead.  They exist here
 * either for technical reasons or for when there is no standard library available.
  */

#ifndef LIB_LOG
#define LIB_LOG

#include <memory>
#include <fstream>
#include <ctime>
#include <map>
#include <string>
#include "memory.h"
#include "lockable.h"
#include "singleton.h"

namespace lib 
{
	class log
	{
		// types
	public:
		class line : private lib::lock
		{
			// construction
		public:
			line(log& aParent, bool aDisplayTimestamp, bool aDisplayThreadId) : lib::lock(aParent.iLock), iParent(aParent), iActive(true)
			{
				bool prefixed = false;
				if (aDisplayTimestamp)
				{
					prefixed = true;
					char timestamp[9]; // "hh:mm:ss"
					std::time_t now = std::time(0);
					std::strftime(timestamp, sizeof(timestamp), "%H:%M:%S", std::localtime(&now));
					iParent.iStream << timestamp;
				}
				if (aDisplayThreadId)
				{
					if (prefixed)
						iParent.iStream << " ";
					iParent.iStream << "(t" << ::GetCurrentThreadId() << ")";
					prefixed = true;
				}
				if (prefixed)
					iParent.iStream << ": ";
			}
			line(const line& aOther) : lib::lock(aOther.iParent.iLock), iParent(aOther.iParent), iActive(true)
			{
				aOther.iActive = false;
			}
			~line()
			{
				if (iActive)
					iParent.iStream << std::endl;
			}
			// operations
		public:
			template <typename T>
			line& operator<<(const T& aItem)
			{
				iParent.iStream << aItem;
				return *this;
			}
			// attributes
		private:
			log& iParent;
			mutable bool iActive;
		};
		// construction
	public:
		log(const std::string& aPath) : iStream(aPath.c_str(), std::ios_base::out | std::ios_base::trunc) {}
		// attributes
	private:
		std::ofstream iStream;
		lib::lockable iLock;
	};

	class logger : public lib::singleton<logger>
	{
		// types
	private:
		typedef std::tr1::shared_ptr<lib::log> log_ptr;
		typedef std::map<std::string, log_ptr> container_type;
		// construction
	public:
		logger() : iDirectory("c:/tmp") {}
		lib::log::line log(const std::string& aName, bool aDisplayTimestamp, bool aDisplayThreadId) 
		{
			lib::lock theLock(iLock);
			std::string path = iDirectory + "/" + aName + ".log";
			container_type::iterator i = iLogs.find(path);
			lib::log& theLog = i != iLogs.end() ? *i->second : *iLogs.insert(std::make_pair(path, log_ptr(new lib::log(path)))).first->second;
			return lib::log::line(theLog, aDisplayTimestamp, aDisplayThreadId);
		}
		void set_directory(const std::string& aDirectory) { iDirectory = aDirectory; }
		// attributes
	private:
		std::string iDirectory;
		container_type iLogs;
		lib::lockable iLock;
	};

	struct nologger
	{
		template <typename T>
		nologger& operator<<(const T&) { return *this; }
	};
}

#ifdef ENABLE_LOGGING
	#define LOG_DIRECTORY(a) lib::logger::instance().set_directory(a);
	#ifndef LOG_TIMESTAMP
		#define LOG_TIMESTAMP 1
	#endif
	#ifndef LOG_THREAD_ID
		#define LOG_THREAD_ID 0
	#endif
	#if LOG_TIMESTAMP
		#if LOG_THREAD_ID
			#define LOG(name) lib::logger::instance().log(name, true, true)
		#else
			#define LOG(name) lib::logger::instance().log(name, true, false)
		#endif
	#else 
		#if LOG_THREAD_ID
			#define LOG(name) lib::logger::instance().log(name, false, true)
		#else
			#define LOG(name) lib::logger::instance().log(name, false, false)
		#endif
	#endif
#else
	#define LOG_DIRECTORY(a)
	#define LOG(name) lib::nologger()
#endif

#endif // LIB_LOG
