// map.h
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* WARNING: This class is not a substitute for the std::map available on your platform
 * which you should be using instead.  It is here either for technical reasons or for when 
 * there is no standard library available.
  */

#ifndef LIB_MAP
#define LIB_MAP

#include <memory>
#include "utility.h"
#include "functional.h"
#include "detail_tree.h"

namespace lib 
{
	template<typename Key, typename T, typename Compare = less<Key>, typename A = std::allocator<pair<const Key, T> > >
	class map
	{
	public:
		typedef Key key_type;
		typedef T mapped_type;
		typedef pair<const Key, T> value_type;
	private:
		struct KeyFunc
		{
			const Key& operator()(const value_type& value) const { return value.first; }
		};
		typedef detail::redblack_tree<false, Key, value_type, KeyFunc, Compare, A> tree_impl;
	public:
		typedef typename tree_impl::pointer pointer;
		typedef typename tree_impl::const_pointer const_pointer;
		typedef typename tree_impl::reference reference;
		typedef typename tree_impl::const_reference const_reference;
		typedef typename tree_impl::allocator_type allocator_type;
		typedef typename tree_impl::iterator iterator;
		typedef typename tree_impl::const_iterator const_iterator;

		enum { node_size = tree_impl::node_size };

	private:
		tree_impl iTree;

	public:
		// construction
		map(const A& allocator = A()) : iTree(allocator) {}
		~map() {}

		// traversals
		iterator begin() { return iTree.begin(); }
		const_iterator begin() const { return iTree.begin(); }
		iterator end() { return iTree.end(); }
		const_iterator end() const { return iTree.end(); }
		iterator lower_bound(const key_type& x) { return iTree.lower_bound(x); }
		const_iterator lower_bound(const key_type& x) const { return iTree.lower_bound(x); }
		iterator upper_bound(const key_type& x) { return iTree.upper_bound(x); }
		const_iterator upper_bound(const key_type& x) const { return iTree.upper_bound(x); }
		bool empty() const { return iTree.empty(); }
		size_t tree_depth() const { return iTree.depth(); }
		// element access
		T& operator[](const key_type& x) { return (*((insert(make_pair(x, T()))).first)).second; }
		// modifiers
		pair<iterator, bool> insert(const value_type& value) { return iTree.insert(value); }
		void erase(iterator position) { iTree.erase(position); }
		size_t erase(const key_type& x) 
		{ 
			iterator first = iTree.lower_bound(x);
			iterator last = iTree.upper_bound(x);
			size_t count = 0;
			while(first != last)
				erase(first++), count++;
			return count;
		}
		void erase(iterator first, iterator last) { while(first != last) erase(first++); }
		void remove(const T& value)
		{
			for (iterator i = begin(); i != end(); )
			{
				if (*i == value)
					erase(i++);
				else
					i++;
			}
		}
	};
}

#endif // LIB_MAP