// network.h - v2.5.1
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef LIB_NETWORK
#define LIB_NETWORK

#include <ws2tcpip.h>
#include <wspiapi.h>
#undef min
#undef max
#include <vector>
#include <list>
#include <map>
#ifndef NO_SSL
#include <openssl/ssl.h>
#endif //NO_SSL
#include "threadable.h"
#include "event.h"
#include "observable.h"
#include "variant.h"

namespace lib
{
	enum { MaxPacketLength = 10240 };
	enum { PacketHeaderLength = sizeof(int) + sizeof(size_t) };
	enum { SocketBufferLength = MaxPacketLength + PacketHeaderLength };

	class packet
	{
	public:
		typedef std::vector<char> buffer_t;
		// construction
	public:
		packet();
		packet(const packet& aPacket);
		packet(int aId);
		packet(int aId, size_t aLength, const void* aData);
		packet(int aId, char aData);
		packet(int aId, int aData);
		packet(size_t aLength, const void* aData);
		~packet();
		// operations
	public:
		void set(int aId, size_t aLength, const void* aData);
		void set(int aId, char aData);
		void set(int aId, int aData);
		void set(size_t aLength, const void* aData);
		int id() const { return iId; }
		size_t length() const { return iData.size(); }
		const void* data() const { return iData.empty() ? 0 : &iData[0]; }
		void* data() { return iData.empty() ? 0 : &iData[0]; }
		const buffer_t& buffer() const { return iData; }
		buffer_t& buffer() { return iData; }
		// implementation
	private:
		void swap(packet& aOther);
		// attributes
	private:
		int iId;
		buffer_t iData;
	};

	// WARNING: Think about endianness if using packet_value<T>
	template <typename T>
	T packet_value(const packet& aPacket)
	{
		assert(aPacket.length() == sizeof(T));
		return *reinterpret_cast<const T*>(aPacket.data());
	}
	template <>
	std::string packet_value<std::string>(const packet& aPacket);

	class serializable;
	void save_to_packet(packet& aPacket, serializable& aObject);
	void load_from_packet(const packet& aPacket, serializable& aObject);

	typedef std::list<packet> packet_list;

	class socket
	{
		friend class socket_manager; 
		// types
	public:
		enum blocking_type { Blocking, NonBlocking };
		enum protocol_type { Packet, Basic };
	protected:
		enum status { StatusNone = 0x00, StatusRead = 0x01, StatusWrite = 0x02, StatusExcept = 0x04 };
		// construction
	public:
		socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol);
		socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol, SOCKET aHandle, bool aConnected = false);
		socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol, SOCKET aHandle, u_long aAddress, u_short aPort, bool aConnected = false);
		socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol, u_long aAddress, u_short aPort, bool aConnect = true);
		virtual ~socket();
		// operations
	public:
		socket_manager& owner() { return iOwner; }
		bool create(bool aCoalesce = false);
		bool connect(bool aCoalesce = false);
		bool connected() const { return iConnected; }
		void close();
		SOCKET handle() const { return iHandle; }
		u_long address() const { return iAddress; }
		u_long local_address() const;
		bool set_blocking(blocking_type aBlocking);
		bool set_blocking() { return set_blocking(iBlocking); }
		blocking_type blocking() const { return iBlocking; }
		void set_protocol(protocol_type aProtocol);
		protocol_type protocol() const { return iProtocol; }
		void set_has_underflow(bool aHasUnderflow) { iHasUnderflow = aHasUnderflow; }
		bool has_underflow() const { return iHasUnderflow; }
		bool underflow() const { return iUnderflow; }
		void output(const packet& aOutputPacket);
		u_short port() const { return iPort; }
		u_short local_port() const;
		socket* parent() const { return iParent; }
		void set_parent(socket* aParent) { iParent = aParent; }
		bool zombie() const { return iHandle == INVALID_SOCKET; }
		bool has_error() const { return iError != 0; }
		int error() const { return iError; }
		const std::string& error_string() const { return iErrorString; }
		// implementation
	protected:
		virtual bool has_output() const;
		status get_status() const { return iStatus; }
		void set_status(status aToSet) { iStatus = static_cast<status>(iStatus | aToSet); }
		void clear_status(status aToClear) { iStatus = static_cast<status>(iStatus & ~aToClear); }
		void set_connected(bool aConnected = true);
		virtual void handle_connected() {}
		virtual void handle_disconnected() {}
		virtual bool handle_input();
		virtual bool handle_output();
		virtual bool handle_error();
		virtual bool read(char* aBuffer, int aLength, int& aBytesRead);
		virtual bool read_more() const { return false; }
		virtual bool write(const char* aBuffer, int aLength, int& aBytesWritten);
		enum retry_op_e { None, Read, Write };
		virtual bool try_read_again(retry_op_e aRetryOp) const { return false; }
		virtual bool try_write_again(retry_op_e aRetryOp) const { return false; }
		// implementation
	protected:
		void get_error();
		void set_error(int aError);
		// attributes
	protected:
		socket_manager& iOwner;
		bool iSocketIsBlocking;
		blocking_type iBlocking;
		protocol_type iProtocol;
		SOCKET iHandle;
		socket* iParent;
		int iError;
		std::string iErrorString;
		u_long iAddress;
		u_short iPort;
		bool iConnected;
		char iInputBuffer[SocketBufferLength];
		char* iInputBufferRecvPtr;
		char* iInputBufferNextPacketPtr;
		char iOutputBuffer[SocketBufferLength];
		char* iOutputBufferNextPacketPtr;
		packet_list iInputPackets;
		packet_list iOutputPackets;
		packet iBasicProtocolPacket;
		status iStatus;
		bool iHasUnderflow;
		bool iUnderflow;
	};

	class listen_socket : public socket
	{
		// construction
	public:
		listen_socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol, u_short aPort);
		// operations
	public:
		bool listen();
		// implementation
	private:
		virtual bool handle_input();
		virtual bool handle_output() { return true; }
	};

	#ifndef NO_SSL
	class ssl_socket : public socket
	{
		// construction
	public:
		ssl_socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol);
		ssl_socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol, SOCKET aHandle, bool aConnected = false);
		ssl_socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol, SOCKET aHandle, u_long aAddress, u_short aPort, bool aConnected = false);
		ssl_socket(socket_manager& aOwner, blocking_type aBlocking, protocol_type aProtocol, u_long aAddress, u_short aPort, bool aConnect = true);
		virtual ~ssl_socket();
		// implementation
	private:
		bool initializing() const;
		bool pending() const;
		virtual bool has_output() const;
		virtual void handle_connected();
		virtual void handle_disconnected();
		virtual bool read(char* aBuffer, int aLength, int& aBytesRead);
		virtual bool read_more() const;
		virtual bool write(const char* aBuffer, int aLength, int& aBytesWritten);
		virtual bool try_read_again(retry_op_e aRetryOp) const { return iReadSsl == aRetryOp; }
		virtual bool try_write_again(retry_op_e aRetryOp) const { return iWriteSsl == aRetryOp; }
		// attributes
	private:
		SSL* iSsl;
		BIO* iBio;
		retry_op_e iReadSsl;
		retry_op_e iWriteSsl;
	};
	#endif //NO_SSL

	class resolver
	{
		// types
	public:
		struct requester
		{
			virtual void host_resolved(const std::string& aHostName, const std::string& aHostNumber, u_long aAddress) = 0;
		};
		struct entry : public threadable
		{
			resolver& iParent;
			const socket* iSocket;
			requester* iRequester;
			std::string iHostName;
			std::string iHostNumber;
			bool iDoReverseLookup;
			u_long iAddress;
			bool iReady;
			int iError;
			entry(resolver& aParent) : iParent(aParent), iSocket(0), iRequester(0), iAddress(INADDR_NONE), iReady(false), iError(0) {}
			entry(resolver& aParent, const socket& aSocket) : iParent(aParent), iSocket(&aSocket), iRequester(0), iAddress(INADDR_NONE), iReady(false), iError(0) {}
			entry(resolver& aParent, requester& aRequester) : iParent(aParent), iSocket(0), iRequester(&aRequester), iAddress(INADDR_NONE), iReady(false), iError(0) {}
			entry& operator=(const entry& aOther);
			virtual void task();
			virtual bool soft_abort() const { return false; }
		};
		typedef std::list<entry> entries;
		// construction
	public:
		resolver();
		~resolver();
		// operations
	public:
		void add(const std::string& aHostName, bool aDoReverseLookup = false);
		void add(const socket& aSocket, const std::string& aHostName, bool aDoReverseLookup = false);
		void add(requester& aRequester, const std::string& aHostName, bool aDoReverseLookup = false);
		void remove(const socket& aSocket);
		void remove(requester& aRequester);
		bool is_next_ready() const;
		bool next_ready(entry& aEntry);
		// attributes
	private:
		entries iEntries;
	};

	class processor
	{
		// construction
	public:
		processor(socket_manager& aOwner);
		~processor();
		// operations
	public:
		socket_manager& owner() const { return iOwner; }
		void task();
		void process();
		bool wait(DWORD aTimeOut);
		// attributes
	private:
		socket_manager& iOwner;
		HWND iSink;
		lib::event iEvent;
		bool iHasWork;
	};

	class socket_manager_observer
	{
		friend class socket_manager;
	private:
		virtual void socket_added(socket& aSocket, bool aListener) = 0;
		virtual void socket_removed(socket& aSocket) = 0;
		virtual void socket_connected(socket& aSocket) = 0;
		virtual void socket_data(socket& aSocket, const packet& aPacket) = 0;
		virtual void socket_underflow(socket& aSocket) = 0;
		virtual void host_resolved(const std::string& aHostName, const std::string& aHostNumber, u_long aAddress) = 0;
	public:
		enum notify_type { NotifySocketAdded, NotifySocketRemoved, NotifySocketConnected, NotifyData, NotifyUnderflow, NotifyHostResolved };
	};

	class socket_manager : public observable<socket_manager_observer>, public threadable
	{
		friend class socket;
		friend class listen_socket;
		friend class ssl_socket;
		friend class processor;
		// types
	public:
		enum processor_type_e { Internal, External };
		typedef bool (*yield_proc)();
		// construction
	public:
		socket_manager(void* aApplicationHandle, processor_type_e aProcessorType = Internal, socket::blocking_type aBlocking = socket::NonBlocking, socket::protocol_type aProtcol = socket::Basic);
		~socket_manager();
		// operations
	public:
		void* application_handle() const { return iApplicationHandle; }
		processor_type_e processor_type() const { return iProcessorType; }
		void process_any(DWORD aTimeOut = 20);
		void set_yield_proc(yield_proc aYieldProc) { iYieldProc = aYieldProc; }
		bool yield() const { return iYieldProc && iYieldProc(); }
		socket* listen(u_short aPort);
		socket* connect(u_long aAddress, u_short aPort, bool aSecure = false);
		socket* connect(const std::string& aHostName, u_short aPort, bool aSecure = false);
		bool empty() const { return iSockets.empty(); }
		bool connected() const;
		void disconnect();
		void disconnect(socket& aSocket);
		bool has_error() const { return iError != 0; }
		int error() const { return iError; }
		std::string error_string() const { return iErrorString; }
		lib::resolver& resolver() { return iResolver; }
		using threadable::lock;
		using threadable::unlock;
		// implementation
	private:
		void process();
		#ifndef NO_SSL
		void close_ssl();
		#endif
		bool do_select(fd_set& aReadFds, fd_set& aWriteFds, fd_set& aExceptFds);
		void do_io();
		typedef variant<packet, bool, void*> notification_data;
		void add_notification(socket_manager_observer::notify_type aNotificationType, socket& aSocket, const notification_data& aData = (void*)0);
		void process_notifications();
		void process_zombies();
		void process_resolved();
		void get_error();
		void clear_error();
		u_long to_address(const std::string& aHostName);
		void add_socket(socket* aSocket, bool aListener = false);
		void remove_socket(socket* aSocket);
		#ifndef NO_SSL
		void start_ssl();
		SSL_CTX* ssl_ctx();
		#endif
		// from observable<socket_manager_observer>
		virtual void notify_observer(socket_manager_observer& aObserver, socket_manager_observer::notify_type aType, const void* aParameter = 0, const void* aParameter2 = 0);
		// from threadable
		virtual void task();
		virtual bool soft_abort() const { return true; }		
		// attributes
	private:
		void* iApplicationHandle;
		processor_type_e iProcessorType;
		processor iProcessor;
		yield_proc iYieldProc;
		bool iProcessing;
		socket::protocol_type iProtocol;
		socket::blocking_type iBlocking;
		int iError;
		std::string iErrorString;
		bool iInitialised;
		lib::resolver iResolver;
		typedef std::list<socket*> socket_list;
		socket_list iSockets;
		typedef std::map<socket*, std::pair<bool, SOCKET> > fd_list;
		fd_list iReadFds;
		fd_list iWriteFds;
		fd_list iExceptFds;
		struct notification_entry
		{
			socket_manager_observer::notify_type iType;
			socket& iSocket;
			notification_data iData;
		};
		typedef std::list<notification_entry> notification_entries;
		notification_entries iNotificationEntries;
		#ifndef NO_SSL
		bool iSslStarted;
		SSL_METHOD* iSslMethod;
		SSL_CTX* iSslCtx;
		#endif //NO_SSL
	};
}

#endif // LIB_NETWORK