// observable.h v1.2
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef LIB_OBSERVABLE
#define LIB_OBSERVABLE

#include <list>
#include <algorithm>
#include <cassert>

namespace lib
{
	template <typename Observer>
	class observable
	{
		// types
	public:
		typedef Observer observer_type;
		typedef typename observer_type::notify_type notify_type;

		// construction
	public:
		observable() {}
		observable(const observable&) {} // no op == no copying of observer lists
		observable& operator=(const observable&) { return *this; } // no op == no copying of observer lists
		virtual ~observable()
		{
			for (task_list::iterator i = iTasks.begin(); i != iTasks.end(); ++i)
				**i = false;
		}

		// operations
	public:
		virtual void add_observer(observer_type& aObserver)
		{
			iObservers.push_back(&aObserver);
		}
		virtual void remove_observer(observer_type& aObserver)
		{
			iObservers.remove(&aObserver);
			for (typename notification_list::iterator i = iNotifications.begin(); i != iNotifications.end(); ++i)
				i->remove(&aObserver);
		}
		template <typename T>
		void notify_observers(notify_type aType, const T& aParameter)
		{
			notify_observers(aType, static_cast<const void*>(&aParameter));
		}
		template <typename T1, typename T2>
		void notify_observers(notify_type aType, const T1& aParameter, const T2& aParameter2)
		{
			notify_observers(aType, static_cast<const void*>(&aParameter), static_cast<const void*>(&aParameter2));
		}
		void notify_observers(notify_type aType, const void* aParameter = 0, const void* aParameter2 = 0)
		{
			bool task = true;
			iTasks.push_back(&task);
			iNotifications.push_front(iObservers);
			typename notification_list::iterator theNotifications = iNotifications.begin();
			while(!theNotifications->empty())
			{
				observer_type* nextObserver = theNotifications->front();
				theNotifications->pop_front();
				notify_observer(*nextObserver, aType, aParameter, aParameter2);
				if (!task)
					return;
			}
			iNotifications.erase(theNotifications);
			iTasks.remove(&task);
		}
		template <typename T>
		void notify_observer(observer_type& aObserver, notify_type aType, const T& aParameter)
		{
			notify_observer(aObserver, aType, static_cast<const void*>(&aParameter));
		}
		template <typename T1, typename T2>
		void notify_observer(observer_type& aObserver, notify_type aType, const T1& aParameter, const T2& aParameter2)
		{
			notify_observer(aObserver, aType, static_cast<const void*>(&aParameter), static_cast<const void*>(&aParameter2));
		}

		// implementation
	protected:
		virtual void notify_observer(observer_type& aObserver, notify_type aType, const void* aParameter = 0, const void* aParameter2 = 0) = 0;

		// attributes
	protected:
		typedef std::list<observer_type*> observer_list;
		observer_list iObservers;
		typedef std::list<observer_list> notification_list;
		notification_list iNotifications;
		typedef std::list<bool*> task_list;
		task_list iTasks;
	};
}

#endif //LIB_OBSERVABLE
