// plugin_impl.h
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef LIB_PLUGIN_IMPL
#define LIB_PLUGIN_IMPL

#include "string.h"
#include "variant.h"
#include "lockable.h"
#include "event.h"
#include "plugin.h"

namespace lib
{
	namespace plugin_impl
	{
		class string : public plugin::string
		{
		public:
			string(const lib::lockable& aLock) : iLock(aLock), iRefCount(0) {}
			string(const lib::lockable& aLock, const std::string& aContents) : iLock(aLock), iContents(aContents), iRefCount(0) {}
			string(const lib::lockable& aLock, const std::wstring& aContents) : iLock(aLock), iContents(aContents), iRefCount(0) {}
		public:
			virtual void add_ref() const;
			virtual void release() const;	
			virtual const_utf8_pointer as_utf8() const;
			virtual const_utf16_pointer as_utf16() const;
			virtual string& operator=(const_utf8_pointer aUtf8String);
			virtual string& operator=(const_utf16_pointer aUtf16String);
			bool valid() const { return iContents.valid(); }
			void clear() { iContents.clear(); }
		private:
			mutable lib::variant<std::string, std::wstring> iContents;
			mutable std::size_t iRefCount;
			const lib::lockable& iLock;
		};

		class string_holder
		{
			// types
		private:
			typedef const string* const_pointer;
			typedef string* pointer;
			typedef const string& const_reference;
			typedef string& reference;
			// construction
		public:
			string_holder() { add_ref(); }
			string_holder(const plugin::string* aString) : iString(static_cast<const_pointer>(aString)) { add_ref(); }
			string_holder(plugin::string* aString) : iString(static_cast<pointer>(aString)) { add_ref(); }
			string_holder(const string_holder& aOther) : iString(aOther.iString) { add_ref(); }
			~string_holder() { release(); }
			string_holder& operator=(const string_holder& aOther) { release(); iString = aOther.iString; add_ref(); return *this; }
			// operations
		public:
			bool valid() const 
			{ 
				return iString.valid() && operator*().valid(); 
			}
			void clear() 
			{
				if (iString.valid())
					operator*().clear(); 
			}
			const_pointer operator->() const 
			{ 
				if (iString.is<const_pointer>())
					return static_cast<const_pointer>(iString); 
				else if (iString.is<pointer>())
					return static_cast<pointer>(iString); 
				else
					return 0;
			}
			pointer operator->()
			{ 
				if (iString.is<pointer>())
					return static_cast<pointer>(iString); 
				else if (iString.is<const_pointer>())
					throw std::logic_error("string_holder::operator->");
				else
					return 0;
			}
			const_reference operator*() const 
			{ 
				if (iString.valid())
					return *operator->(); 
				else
					throw std::logic_error("string_holder::operator*");
			}
			reference operator*()
			{ 
				if (iString.valid())
					return *operator->(); 
				else
					throw std::logic_error("string_holder::operator*");
			}
			// implementation
		private:
			void add_ref() const
			{
				if (iString.is<const_pointer>())
					return static_cast<const_pointer>(iString)->add_ref();
				else if (iString.is<pointer>())
					return static_cast<pointer>(iString)->add_ref();
			}
			void release() const
			{
				if (iString.is<const_pointer>())
					return static_cast<const_pointer>(iString)->release();
				else if (iString.is<pointer>())
					return static_cast<pointer>(iString)->release();
			}
			// attributes
		private:
			lib::variant<pointer, const_pointer> iString;
		};

		class result : public plugin::object
		{
		public:
			typedef lib::variant<bool> contents_t;
		public:
			result(lib::lockable& aLock) : iLock(aLock), iRefCount(0) {}
		public:
			template <typename T>
			operator const T&() const { return static_cast<const T&>(iContents); }
			template <typename T>
			operator T&() { return static_cast<T&>(iContents); }
			const contents_t& contents() const { return iContents; }
			contents_t& contents() { return iContents; }
			virtual void add_ref() const { lib::lock lock(iLock); ++iRefCount; }
			virtual void release() const { lib::lock lock(iLock); if (--iRefCount == 0) delete this; }
		private:
			lib::lockable& iLock;
			mutable std::size_t iRefCount;
			contents_t iContents;
		};

		class result_holder
		{
			// types
		private:
			typedef const result* const_pointer;
			typedef result* pointer;
			typedef const result& const_reference;
			typedef result& reference;
			// construction
		public:
			result_holder() { add_ref(); }
			result_holder(const_pointer aResult) : iResult(static_cast<const_pointer>(aResult)) { add_ref(); }
			result_holder(pointer aResult) : iResult(static_cast<pointer>(aResult)) { add_ref(); }
			result_holder(const result_holder& aOther) : iResult(aOther.iResult) { add_ref(); }
			~result_holder() { release(); }
			result_holder& operator=(const result_holder& aOther) { release(); iResult = aOther.iResult; add_ref(); return *this; }
			// operations
		public:
			bool valid() const
			{
				return iResult.valid();
			}
			const_pointer operator->() const 
			{ 
				if (iResult.is<const_pointer>())
					return static_cast<const_pointer>(iResult); 
				else if (iResult.is<pointer>())
					return static_cast<pointer>(iResult); 
				else
					return 0;
			}
			pointer operator->()
			{ 
				if (iResult.is<pointer>())
					return static_cast<pointer>(iResult); 
				else if (iResult.is<const_pointer>())
					throw std::logic_error("result_holder::operator->");
				else
					return 0;
			}
			const_reference operator*() const 
			{ 
				if (iResult.valid())
					return *operator->(); 
				else
					throw std::logic_error("result_holder::operator*");
			}
			reference operator*()
			{ 
				if (iResult.valid())
					return *operator->(); 
				else
					throw std::logic_error("result_holder::operator*");
			}
			// implementation
		private:
			void add_ref() const
			{
				if (iResult.is<const_pointer>())
					return static_cast<const_pointer>(iResult)->add_ref();
				else if (iResult.is<pointer>())
					return static_cast<pointer>(iResult)->add_ref();
			}
			void release() const
			{
				if (iResult.is<const_pointer>())
					return static_cast<const_pointer>(iResult)->release();
				else if (iResult.is<pointer>())
					return static_cast<pointer>(iResult)->release();
			}
			// attributes
		private:
			lib::variant<pointer, const_pointer> iResult;
		};

		template <typename SpecificPluginHost>
		class plugin_host : public SpecificPluginHost, public lib::lockable
		{
		public:
			virtual void create_string(plugin::string*& aString) const
			{
				aString = new string(*this);
				aString->add_ref();
			}
			virtual void create_string(plugin::string*& aString, string::const_pointer aContents) const
			{
				aString = new string(*this);
				aString->add_ref();
				*aString = aContents;
			}
			virtual void create_string(plugin::string*& aString, string::const_wide_pointer aContents) const
			{
				aString = new string(*this);
				aString->add_ref();
				*aString = aContents;
			}
		};
	}
}

#endif //LIB_PLUGIN_IMPL
