// queue.h - v1.4
/*
 *  Copyright (c) 2011 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef LIB_QUEUE
#define LIB_QUEUE

#include <list>
#include <vector>
#include "interlockable.h"
#include "event.h"

namespace lib
{
	template <typename QueueItem>
	class queue : public interlockable
	{
		// types
	public:
		typedef QueueItem value_type;
		class sink
		{
		public:
			virtual void from_queue(queue& aQueue, QueueItem& aItem) = 0;
		};
	private:
		class entry
		{
			// types
		public:
			enum type_e { Send, Post };
			// construction
		public:
			entry(const QueueItem& aItem, type_e aType) : iItem(aItem), iType(aType), iEvent(0), iFrom(0) {}
			entry(const QueueItem& aItem, type_e aType, const event& aEvent) : iItem(aItem), iType(aType), iEvent(&aEvent), iFrom(0) {}
			entry(const QueueItem& aItem, type_e aType, const interlockable& aFrom) : iItem(aItem), iType(aType), iEvent(0), iFrom(&aFrom) {}
			entry(const QueueItem& aItem, type_e aType, const event& aEvent, const interlockable& aFrom) : iItem(aItem), iType(aType), iEvent(&aEvent), iFrom(&aFrom) {}
			// operations
		public:
			const QueueItem& item() const { return iItem; }
			QueueItem& item() { return iItem; }
			type_e type() const { return iType; }
			const interlockable* from() const { return iFrom; }
			void signal() const { if (iEvent != 0) iEvent->signal(); }
			// attributes
		private:
			QueueItem iItem;
			type_e iType;
			const event* iEvent;
			const interlockable* iFrom;
		};
		typedef std::list<entry> container_type;
		typedef typename container_type::const_iterator const_iterator;
		typedef typename container_type::iterator iterator;
		typedef std::vector<iterator> working_list;
		// construction
	public:
		queue() :  iSink(0) {}
		queue(sink& aSink) : iSink(&aSink) {}
		// operations
	public:
		bool wait(event::timeout_interval aTimeout = event::TimeoutInfinite)
		{
			return iNewItemEvent.wait(aTimeout) != 0;
		}
		void loop(event::timeout_interval aTimeout = event::TimeoutInfinite, bool (*aYieldProc)() = 0)
		{
			if (iSink == 0)
				throw std::logic_error("lib::queue::loop");
			while ((!aYieldProc || !aYieldProc()) && wait(aTimeout))
				process_queue();
		}
		bool send(const QueueItem& aItem, event::timeout_interval aTimeout = event::TimeoutInfinite)
		{
			return send(iProcessedItemEvent, aItem, aTimeout);
		}
		bool send(const event& aEvent, const QueueItem& aItem, event::timeout_interval aTimeout = event::TimeoutInfinite)
		{
			lock();
			iItems.push_back(entry(aItem, entry::Send, aEvent));
			unlock();
			iNewItemEvent.signal();
			return aEvent.wait(aTimeout) != 0;
		}
		bool interlocked_send(const interlockable& aOther, const QueueItem& aItem, event::timeout_interval aTimeout = event::TimeoutInfinite)
		{
			return interlocked_send(aOther, iProcessedItemEvent, aItem, aTimeout);
		}
		bool interlocked_send(const interlockable& aOther, const event& aEvent, const QueueItem& aItem, event::timeout_interval aTimeout = event::TimeoutInfinite)
		{
			interlock_acquire(aOther);
			lock();
			iItems.push_back(entry(aItem, entry::Send, aEvent, aOther));
			unlock();
			iNewItemEvent.signal();
			interlock_release(aOther);
			return aEvent.wait(aTimeout) != 0;
		}
		void post(const QueueItem& aItem)
		{
			lock();
			iItems.push_back(entry(aItem, entry::Post));
			unlock();
			iNewItemEvent.signal();
		}
		QueueItem& next()
		{
			lib::lock lock(*this);
			iterator nextEntry = next_available();
			iWorkingList.push_back(nextEntry);
			return nextEntry->item();
		}
		void pop_next()
		{
			lib::lock lock(*this);
			iterator nextEntry = iWorkingList.back();
			iWorkingList.pop_back();
			if (nextEntry->type() == entry::Send)
				nextEntry->signal();
			iItems.erase(nextEntry);
		}
		bool any() const
		{
			lib::lock lock(*this);
			return next_available() != iItems.end();
		}
		void clear()
		{
			lib::lock lock(*this);
			iItems.clear();
		}
		// implementation
	private:
		const_iterator next_available() const
		{
			const_iterator nextAvailable;
			if (iWorkingList.empty())
				nextAvailable = iItems.begin();
			else
			{
				nextAvailable = iWorkingList.back();
				++nextAvailable;
			}
			return nextAvailable;
		}
		iterator next_available()
		{
			iterator nextAvailable;
			if (iWorkingList.empty())
				nextAvailable = iItems.begin();
			else
			{
				nextAvailable = iWorkingList.back();
				++nextAvailable;
			}
			return nextAvailable;
		}
		void process_queue()
		{
			assert(iSink != 0);
			while (any())
			{
				QueueItem& nextItem = next();
				iSink->from_queue(*this, nextItem);
				pop_next();
			}
		}
		// from interlockable
		virtual bool purge(const interlockable& aOther)
		{
			if (iSink == 0)
				throw std::logic_error("lib::queue::purge");
			lib::lock lock(*this);
			for (working_list::const_iterator i = iWorkingList.begin(); i != iWorkingList.end(); ++i)
			{
				entry& theEntry = **i;
				if (theEntry.type() == entry::Send && theEntry.from() == &aOther)
					return false;
			}
			if (wait(event::TimeoutShort))
				process_queue();
			return true;
		}
		// attributes
	private:
		container_type iItems;
		working_list iWorkingList;
		event iNewItemEvent;
		event iProcessedItemEvent;
		sink* iSink;
	};
}

#endif //LIB_QUEUE
