// quick_string.h - v1.1
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef LIB_QUICK_STRING
#define LIB_QUICK_STRING

#include <string>
#include <algorithm>
#include <memory>
#include <iterator>
#include <stdexcept>

namespace lib 
{
	template <typename charT, typename Traits = std::char_traits<charT>, typename Alloc = std::allocator<charT> >
	struct quick_string
	{
	public:
		typedef typename std::basic_string<charT, Traits, Alloc> string_type;
		typedef typename string_type::traits_type traits_type;
		typedef typename string_type::allocator_type allocator_type;
		typedef typename string_type::value_type value_type;
		typedef typename string_type::size_type size_type;
		typedef typename string_type::difference_type difference_type;
		typedef typename string_type::reference reference;
		typedef typename string_type::const_reference const_reference;
		typedef typename string_type::pointer pointer;
		typedef typename string_type::const_pointer const_pointer;
		typedef typename string_type::iterator iterator;
		typedef typename string_type::const_iterator const_iterator;
		typedef std::reverse_iterator<iterator> reverse_iterator;
		typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
		static const size_type npos;
	private:
		typedef typename std::pair<typename string_type::const_iterator, typename string_type::const_iterator> remote_string;
	public:
		// construct/copy/destroy
		explicit quick_string(const Alloc& a = Alloc()) : iString(a), iRemote(false) {}
		quick_string(const string_type& str) : iString(str), iRemote(false) {}
		quick_string(const quick_string& str) : iString(str.iString), iRemote(str.iRemote), iRemoteString(str.iRemoteString) {}
		quick_string(const quick_string& str, size_type pos, size_type n = npos, const Alloc& a = Alloc()) : iRemote(false)
		{
			if (!str.is_remote())
				local().assign(static_cast<const string_type&>(str), pos, n);
			else
			{
				iRemote = true;
				iRemoteString = remote_string(static_cast<const remote_string&>(str).first + pos, 
					n == npos ? static_cast<const remote_string&>(str).second : 
					static_cast<const remote_string&>(str).first + pos + n);
			}
		}
		quick_string(const charT* s, size_type n, const Alloc& a = Alloc()) : iString(s, n, a), iRemote(false) {}
		quick_string(const charT* s, const Alloc& a = Alloc()) : iString(s, a), iRemote(false) {}
		quick_string(size_type n, charT c, const Alloc& a = Alloc()) : iString(n, c, a), iRemote(false) {}
		template<class InputIterator>
		quick_string(InputIterator begin, InputIterator end, const Alloc& a = Alloc()) : iString(begin, end, a), iRemote(false) {}
		quick_string(const quick_string& str, const Alloc& a) : iString(str.iString, a), iRemote(str.iRemote), iRemoteString(str.iRemoteString) {}
		quick_string(const_iterator begin, const_iterator end) : iRemote(true), iRemoteString(begin, end) {}
		quick_string(iterator begin, iterator end) : iRemote(true), iRemoteString(begin, end) {}
		quick_string(const remote_string& aRemoteString) : iRemote(true), iRemoteString(aRemoteString) {}
		quick_string& operator=(const quick_string& str) { iString = str.iString; iRemote = str.iRemote; iRemoteString = str.iRemoteString; return *this; }
		quick_string& operator=(const charT* s) { iString = s; iRemote = false; return *this; }
		quick_string& operator=(charT c) { iString = c; iRemote = false; return *this; }
		// iterators
		iterator begin() { return local().begin(); }
		const_iterator begin() const { return is_remote() ? iRemoteString.first : iString.begin(); }
		const_iterator cbegin() const { return is_remote() ? iRemoteString.first : iString.begin(); }
		iterator end() { return local().end(); }
		const_iterator end() const { return is_remote() ? iRemoteString.second : iString.end(); }
		const_iterator cend() const { return is_remote() ? iRemoteString.second : iString.end(); }
		reverse_iterator rbegin() { return reverse_iterator(end()); }
		const_reverse_iterator rbegin() const { return const_reverse_iterator(end()); }
		const_reverse_iterator crbegin() const { return const_reverse_iterator(end()); }
		reverse_iterator rend() { return reverse_iterator(begin()); }
		const_reverse_iterator rend() const { return const_reverse_iterator(begin()); }
		const_reverse_iterator crend() const { return const_reverse_iterator(begin()); }
		// capacity
		size_type size() const { return is_remote() ? iRemoteString.second - iRemoteString.first : iString.size(); }
		size_type length() const { return is_remote() ? iRemoteString.second - iRemoteString.first : iString.length(); }
		size_type max_size() const { return iString.max_size(); }
		void resize(size_type n, charT c) { local().resize(n, c); }
		void resize(size_type n) { local().resize(n); }
		size_type capacity() const { return iString.capacity(); }
		void reserve(size_type res_arg = 0) { local().reserve(res_arg); }
		void shrink_to_fit() { local().shrink_to_fit(); }
		void clear() { local().clear(); }
		bool empty() const { return is_remote() ? iRemoteString.first == iRemoteString.second : iString.empty(); }
		// element access
		const_reference operator[](size_type pos) const 
		{ 
			if (is_remote())
			{
				if (pos == iRemoteString.second - iRemoteString.first)
				{
					static value_type nullTerminator = value_type();
					return nullTerminator;
				}
				else
					return *(iRemoteString.first + pos);
			}
			else
				return iString[pos]; 
		}
		reference operator[](size_type pos)
		{ 
			return local()[pos]; 
		}
		const_reference at(size_type n) const 
		{ 
			if (is_remote()) 
			{ 
				if (n >= iRemoteString.second - iRemoteString.first)
					throw std::out_of_range("quick_string::at");
				else
					return *(iRemoteString.first + n);
			}
			else
				return iString.at(n);
		}
		reference at(size_type n)
		{ 
			return local().at(n);
		}
		// modifiers
		quick_string& operator+=(const string_type& str) { local().operator+=(str); return *this; }
		quick_string& operator+=(const charT* s) { local().operator+=(s); return *this; }
		quick_string& operator+=(charT c) { local().operator+=(c); return *this; }
		quick_string& append(const string_type& str) { local().append(str); return *this; }
		quick_string& append(const string_type& str, size_type pos, size_type n) { local().append(str, pos, n); return *this; }
		quick_string& append(const charT* s, size_type n) { local().append(s, n); return *this; }
		quick_string& append(const charT* s) { local().append(s); return *this; }
		quick_string& append(size_type n, charT c) { local().append(n, c); return *this;  }
		template<class InputIterator>
		quick_string& append(InputIterator first, InputIterator last) { local.append(first, last); return *this; }
		void push_back(charT c) { local().push_back(c); }
		quick_string& assign(const string_type& str) { local().assign(str); return *this; }
		quick_string& assign(const string_type& str, size_type pos, size_type n) { local().assign(str, pos, n); return *this; }
		quick_string& assign(const charT* s, size_type n) { local().assign(s, n); return *this; }
		quick_string& assign(const charT* s) { local().assign(s); return *this; }
		quick_string& assign(size_type n, charT c) { local().assign(n, c); return *this; }
		template<class InputIterator>
		quick_string& assign(InputIterator first, InputIterator last) { local().assign(first, last); return *this; }
		quick_string& insert(size_type pos1, const string_type& str) { local().insert(pos1, str); return *this; }
		quick_string& insert(size_type pos1, const string_type& str, size_type pos2, size_type n) { local().insert(pos1, str, pos2, n); return *this; }
		quick_string& insert(size_type pos, const charT* s, size_type n) { local().insert(pos, s, n); return *this; }
		quick_string& insert(size_type pos, const charT* s) { local().insert(pos, s); return *this; }
		quick_string& insert(size_type pos, size_type n, charT c) { local().insert(pos, n, c); return *this; }
		iterator insert(const_iterator p, charT c) { local().insert(p, c); return *this; }
		void insert(const_iterator p, size_type n, charT c) { local().insert(p, n, c); return *this; }
		template<class InputIterator>
		void insert(const_iterator p, InputIterator first, InputIterator last) { local().insert(p, first, last); return *this; }
		quick_string& erase(size_type pos = 0, size_type n = npos) { local().erase(pos, n); return *this; }
		iterator erase(const_iterator p) { return local().erase(p); } 
		iterator erase(const_iterator first, const_iterator last) { return local().erase(first, last); }
		quick_string& replace(size_type pos1, size_type n1, const string_type& str) { local().replace(pos1, n1, str); return *this; }
		quick_string& replace(size_type pos1, size_type n1, const string_type& str, size_type pos2, size_type n2) { local().replace(pos1, n1, str, pos2, n2); return *this; }
		quick_string& replace(size_type pos, size_type n1, const charT* s, size_type n2) { local().replace(pos, n1, s, n2); return *this; }
		quick_string& replace(size_type pos, size_type n1, const charT* s) { local().replace(pos, n1, s); return *this; }
		quick_string& replace(size_type pos, size_type n1, size_type n2, charT c) { local().replace(pos, n1, n2, c); return *this; }
		quick_string& replace(iterator i1, iterator i2, const string_type& str) { local().replace(i1, i2, str); return *this; }
		quick_string& replace(iterator i1, iterator i2, const charT* s, size_type n) { local().replace(i1, i2, s, n); return *this; }
		quick_string& replace(iterator i1, iterator i2, const charT* s) { local().replace(i1, i2, s); return *this; }
		quick_string& replace(iterator i1, iterator i2, size_type n, charT c) { local().replace(i1, i2, n, c); return *this; }
		template<class InputIterator>
		quick_string& replace(iterator i1, iterator i2, InputIterator j1, InputIterator j2) { local().replace(i1, i2, j1, j2); return *this; }
		size_type copy(charT* s, size_type n, size_type pos = 0) const 
		{
			if (is_remote())
			{
				if (size() < pos)
					throw std::out_of_range("quick_string::copy");
				if (size() - pos < n)
					n = size() - pos;
				std::copy(iRemoteString.first, iRemoteString.first + n, s);
				return n;
			}
			else
				return local().copy(s, n, pos);
		}
		void swap(quick_string& str)
		{
			iString.swap(str.iString);
			std::swap(iRemote, str.iRemote);
			std::swap(iRemoteString, str.iRemoteString);
		}
		// string_type operations:
		const charT* c_str() const { return local().c_str(); } // explicit
		const charT* data() const { return local().data(); }
		allocator_type get_allocator() const { return iString.get_allocator(); }
		size_type find (const string_type& str, size_type pos = 0) const 
		{ 
			if (is_remote())
			{
				if (pos < size())
				{
					const_iterator f = std::search(iRemoteString.first + pos, iRemoteString.second, str.begin(), str.end(), Traits::eq);
					if (f != iRemoteString.second)
						return f - iRemoteString.first;
					else
						return npos;
				}
				else
					return npos;
			}
			else
				return iString.find(str, pos);
		}
		size_type find (const charT* s, size_type pos, size_type n) const
		{ 
			if (is_remote())
			{
				if (pos < size())
				{
					const_iterator f = std::search(iRemoteString.first + pos, iRemoteString.second, s, s+n, Traits::eq);
					if (f != iRemoteString.second)
						return f - iRemoteString.first;
					else
						return npos;
				}
				else
					return npos;
			}
			else
				return iString.find(s, pos, n);
		}
		size_type find (const charT* s, size_type pos = 0) const
		{ 
			if (is_remote())
			{
				if (pos < size())
				{
					string_type tmp(s);
					const_iterator f = std::search(iRemoteString.first + pos, iRemoteString.second, tmp.begin(), tmp.end(), Traits::eq);
					if (f != iRemoteString.second)
						return f - iRemoteString.first;
					else
						return npos;
				}
				else
					return npos;
			}
			else
				return iString.find(s, pos);
		}
		size_type find (charT c, size_type pos = 0) const
		{ 
			if (is_remote())
			{
				if (pos < size())
				{
					const_iterator f = std::search(iRemoteString.first + pos, iRemoteString.second, &c, &c+1, Traits::eq);
					if (f != iRemoteString.second)
						return f - iRemoteString.first;
					else
						return npos;
				}
				else
					return npos;
			}
			else
				return iString.find(c, pos);
		}
		size_type rfind(const string_type& str, size_type pos = npos) const
		{ 
			if (is_remote())
			{
				if (pos == npos || pos < size())
				{
					const iterator e = (pos == npos ? iRemoteString.second : iRemoteString.first + pos + 1);
					const_iterator f = std::find_end(iRemoteString.first, e, str.begin(), str.end(), Traits::eq);
					if (f != e)
						return f - iRemoteString.first;
					else
						return npos;
				}
				else
					return npos;
			}
			else
				return iString.rfind(str, pos);
		}
		size_type rfind(const charT* s, size_type pos, size_type n) const
		{ 
			if (is_remote())
			{
				if (pos == npos || pos < size())
				{
					const iterator e = (pos == npos ? iRemoteString.second : iRemoteString.first + pos + 1);
					const_iterator f = std::find_end(iRemoteString.first, e, s, s+n, Traits::eq);
					if (f != e)
						return f - iRemoteString.first;
					else
						return npos;
				}
				else
					return npos;
			}
			else
				return iString.rfind(s, pos, n);
		}
		size_type rfind(const charT* s, size_type pos = npos) const
		{ 
			if (is_remote())
			{
				if (pos == npos || pos < size())
				{
					string_type tmp(s);
					const iterator e = (pos == npos ? iRemoteString.second : iRemoteString.first + pos + 1);
					const_iterator f = std::find_end(iRemoteString.first, e, tmp.begin(), tmp.end(), Traits::eq);
					if (f != e)
						return f - iRemoteString.first;
					else
						return npos;
				}
				else
					return npos;
			}
			else
				return iString.rfind(s, pos);
		}
		size_type rfind(charT c, size_type pos = npos) const
		{ 
			if (is_remote())
			{
				if (pos == npos || pos < size())
				{
					const iterator e = (pos == npos ? iRemoteString.second : iRemoteString.first + pos + 1);
					const_iterator f = std::find_end(iRemoteString.first, e, &c, &c+1, Traits::eq);
					if (f != e)
						return f - iRemoteString.first;
					else
						return npos;
				}
				else
					return npos;
			}
			else
				return iString.rfind(c, pos);
		}
		size_type find_first_of(const string_type& str, size_type pos = 0) const
		{ 
			if (is_remote())
			{
				if (pos < size())
				{
					const_iterator f = std::find_first_of(iRemoteString.first + pos, iRemoteString.second, str.begin(), str.end(), Traits::eq);
					if (f != iRemoteString.second)
						return f - iRemoteString.first;
					else
						return npos;
				}
				else
					return npos;
			}
			else
				return iString.find_first_of(str, pos);
		}
		size_type find_first_of(const charT* s, size_type pos, size_type n) const
		{ 
			if (is_remote())
			{
				if (pos < size())
				{
					const_iterator f = std::find_first_of(iRemoteString.first + pos, iRemoteString.second, s, s+n, Traits::eq);
					if (f != iRemoteString.second)
						return f - iRemoteString.first;
					else
						return npos;
				}
				else
					return npos;
			}
			else
				return iString.find_first_of(s, pos, n);
		}
		size_type find_first_of(const charT* s, size_type pos = 0) const
		{ 
			if (is_remote())
			{
				if (pos < size())
				{
					string_type tmp(s);
					const_iterator f = std::find_first_of(iRemoteString.first + pos, iRemoteString.second, tmp.begin(), tmp.end(), Traits::eq);
					if (f != iRemoteString.second)
						return f - iRemoteString.first;
					else
						return npos;
				}
				else
					return npos;
			}
			else
				return iString.find_first_of(s, pos);
		}
		size_type find_first_of(charT c, size_type pos = 0) const
		{ 
			if (is_remote())
			{
				if (pos < size())
				{
					const_iterator f = std::find_first_of(iRemoteString.first + pos, iRemoteString.second, &c, &c+1, Traits::eq);
					if (f != iRemoteString.second)
						return f - iRemoteString.first;
					else
						return npos;
				}
				else
					return npos;
			}
			else
				return iString.find_first_of(c, pos);
		}
		size_type find_last_of (const string_type& str, size_type pos = npos) const
		{
			if (is_remote())
			{
				if (!empty())
				{
					if (pos == npos)
						pos = size() - 1;
					const_iterator s = iRemoteString.first + pos + 1;
					while (s != iRemoteString.first)
					{
						--s;
						if (str.find(*s) != npos)
							return s - iRemoteString.first;
					} 
				}
				return npos;
			}
			else
				return iString.find_last_of(str, pos);
		}
		size_type find_last_of (const charT* s, size_type pos, size_type n) const
		{
			if (is_remote())
			{
				if (!empty())
				{
					string_type tmp(s, n);
					if (pos == npos)
						pos = size() - 1;
					const_iterator s = iRemoteString.first + pos + 1;
					while (s != iRemoteString.first)
					{
						--s;
						if (tmp.find(*s) != npos)
							return s - iRemoteString.first;
					} 
				}
				return npos;
			}
			else
				return iString.find_last_of(s, pos, n);
		}
		size_type find_last_of (const charT* s, size_type pos = npos) const
		{
			if (is_remote())
			{
				if (!empty())
				{
					string_type tmp(s);
					if (pos == npos)
						pos = size() - 1;
					const_iterator s = iRemoteString.first + pos + 1;
					while (s != iRemoteString.first)
					{
						--s;
						if (tmp.find(*s) != npos)
							return s - iRemoteString.first;
					} 
				}
				return npos;
			}
			else
				return iString.find_last_of(s, pos);
		}
		size_type find_last_of (charT c, size_type pos = npos) const
		{
			if (is_remote())
			{
				if (!empty())
				{
					if (pos == npos)
						pos = size() - 1;
					const_iterator s = iRemoteString.first + pos + 1;
					while (s != iRemoteString.first)
					{
						--s;
						if (Traits::eq(*s, c))
							return s - iRemoteString.first;
					} 
				}
				return npos;
			}
			else
				return iString.find_last_of(c, pos);
		}
		size_type find_first_not_of(const string_type& str, size_type pos = 0) const
		{ 
			if (is_remote())
			{
				if (pos < size())
					for (const_iterator f = iRemoteString.first; f != iRemoteString.second; ++f)
						if (str.find(*f) == npos)
							return f - iRemoteString.first;
				return npos;
			}
			else
				return iString.find_first_not_of(str, pos);
		}
		size_type find_first_not_of(const charT* s, size_type pos, size_type n) const
		{ 
			if (is_remote())
			{
				if (pos < size())
					for (const_iterator f = iRemoteString.first; f != iRemoteString.second; ++f)
						if (str.find(*f) == npos)
							return f - iRemoteString.first;
				return npos;
			}
			else
				return iString.find_first_not_of(str, pos);
		}
		size_type find_first_not_of(const charT* s, size_type pos = 0) const
		{ 
			if (is_remote())
			{
				if (pos < size())
					for (const_iterator f = iRemoteString.first; f != iRemoteString.second; ++f)
						if (str.find(*f) == npos)
							return f - iRemoteString.first;
				return npos;
			}
			else
				return iString.find_first_not_of(str, pos);
		}
		size_type find_first_not_of(charT c, size_type pos = 0) const
		{ 
			if (is_remote())
			{
				if (pos < size())
					for (const_iterator f = iRemoteString.first; f != iRemoteString.second; ++f)
						if (!Traits::eq(*f, c))
							return f - iRemoteString.first;
				return npos;
			}
			else
				return iString.find_first_not_of(c, pos);
		}
		size_type find_last_not_of (const string_type& str, size_type pos = npos) const
		{
			if (is_remote())
			{
				if (!empty())
				{
					if (pos == npos)
						pos = size() - 1;
					const_iterator s = iRemoteString.first + pos + 1;
					while (s != iRemoteString.first)
					{
						--s;
						if (str.find(*s) == npos)
							return s - iRemoteString.first;
					} 
				}
				return npos;
			}
			else
				return iString.find_last_of(str, pos);
		}
		size_type find_last_not_of (const charT* s, size_type pos, size_type n) const
		{
			if (is_remote())
			{
				if (!empty())
				{
					string_type tmp(s, n);
					if (pos == npos)
						pos = size() - 1;
					const_iterator s = iRemoteString.first + pos + 1;
					while (s != iRemoteString.first)
					{
						--s;
						if (tmp.find(*s) == npos)
							return s - iRemoteString.first;
					} 
				}
				return npos;
			}
			else
				return iString.find_last_of(s, pos, n);
		}
		size_type find_last_not_of (const charT* s, size_type pos = npos) const
		{
			if (is_remote())
			{
				if (!empty())
				{
					string_type tmp(s);
					if (pos == npos)
						pos = size() - 1;
					const_iterator s = iRemoteString.first + pos + 1;
					while (s != iRemoteString.first)
					{
						--s;
						if (tmp.find(*s) == npos)
							return s - iRemoteString.first;
					} 
				}
				return npos;
			}
			else
				return iString.find_last_of(s, pos);
		}
		size_type find_last_not_of (charT c, size_type pos = npos) const
		{
			if (is_remote())
			{
				if (!empty())
				{
					if (pos == npos)
						pos = size() - 1;
					const_iterator s = iRemoteString.first + pos + 1;
					while (s != iRemoteString.first)
					{
						--s;
						if (!Traits::eq(*s, c))
							return s - iRemoteString.first;
					} 
				}
				return npos;
			}
			else
				return iString.find_last_of(c, pos);
		}
		string_type substr(size_type pos = 0, size_type n = npos) const
		{
			if (is_remote())
			{
				if (size() < pos)
					throw std::out_of_range("quick_string::copy");
				if (n == npos)
					n = size() - pos;
				else if (size() - pos < n)
					n = size() - pos;
				return string_type(iRemoteString.first + pos, iRemoteString.first + pos + n);
			}
			else
				return iString.substr(pos, n);
		}
		int compare(const quick_string& str) const
		{
			if (is_remote())
			{
				int result;
				if (!empty() && !str.empty())
				{
					if (str.is_remote())
						result = Traits::compare(&*iRemoteString.first, &*str.iRemoteString.first, std::min(size(), str.size()));
					else
						result = Traits::compare(&*iRemoteString.first, str.data(), std::min(size(), str.size()));
				}
				else
					result = 0;
				return (result != 0 ? result : size() < str.size() ? -1 : str.size() == size() ? 0 : +1);
			}
			else
				return iString.compare(str);
		}
		int compare(const string_type& str) const
		{
			if (is_remote())
			{
				int result;
				if (!empty() && !str.empty())
					result = Traits::compare(&*iRemoteString.first, str.data(), std::min(size(), str.size()));
				else
					result = 0;
				return (result != 0 ? result : size() < str.size() ? -1 : str.size() == size() ? 0 : +1);
			}
			else
				return iString.compare(str);
		}
		int compare(size_type pos1, size_type n1, const string_type& str) const
		{
			return quick_string(*this, pos1, n1).compare(str);
		}
		int compare(size_type pos1, size_type n1, const string_type& str, size_type pos2, size_type n2) const
		{
			return quick_string(*this, pos1, n1).compare(quick_string(quick_string(str.begin(), str.end()), pos2, n2));
		}
		int compare(const charT* s) const
		{
			return compare(quick_string(s));
		}
		int compare(size_type pos1, size_type n1, const charT* s) const
		{
			return quick_string(*this, pos1, n1).compare(quick_string(s));
		}
		int compare(size_type pos1, size_type n1, const charT* s, size_type n2) const
		{
			return quick_string(*this, pos1, n1).compare(quick_string(s, n2));
		}
	public:
		// quick_string specific operations
		bool is_remote() const { return iRemote; }
		operator string_type&()
		{
			if (!is_remote())
				return iString;
			else
			{
				iString = string_type(iRemoteString.first, iRemoteString.second);
				iRemote = false;
				return iString;
			}
		}
		operator const string_type&() const
		{
			if (!is_remote())
				return iString;
			else
			{
				iString = string_type(iRemoteString.first, iRemoteString.second);
				iRemote = false;
				return iString;
			}
		}
		operator remote_string&()
		{
			if (is_remote())
				return iRemoteString;
			else
				throw std::logic_error("quick_string::remote_string");
		}
		operator const remote_string&() const
		{
			if (is_remote())
				return iRemoteString;
			else
				throw std::logic_error("quick_string::remote_string");
		}
	private:
		string_type& local() { return static_cast<string_type&>(*this); }
		const string_type& local() const { return static_cast<const string_type&>(*this); }
	private:
		mutable string_type iString;
		mutable bool iRemote;
		remote_string iRemoteString;
	};

	template <typename charT, typename Traits, typename Alloc>
	const typename quick_string<charT, Traits, Alloc>::size_type quick_string<charT, Traits, Alloc>::npos = quick_string<charT, Traits, Alloc>::string_type::npos;


template<class charT,
	class Traits,
	class Alloc> inline
	quick_string<charT, Traits, Alloc> operator+(
		const quick_string<charT, Traits, Alloc>& _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// return quick_string + quick_string
	return (quick_string<charT, Traits, Alloc>(_Left) += _Right);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	quick_string<charT, Traits, Alloc> operator+(
		const charT *_Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// return NTCS + quick_string
	return (quick_string<charT, Traits, Alloc>(_Left) += _Right);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	quick_string<charT, Traits, Alloc> operator+(
		const charT _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// return character + quick_string
	return (quick_string<charT, Traits, Alloc>(1, _Left) += _Right);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	quick_string<charT, Traits, Alloc> operator+(
		const quick_string<charT, Traits, Alloc>& _Left,
		const charT *_Right)
	{	// return quick_string + NTCS
	return (quick_string<charT, Traits, Alloc>(_Left) += _Right);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	quick_string<charT, Traits, Alloc> operator+(
		const quick_string<charT, Traits, Alloc>& _Left,
		const charT _Right)
	{	// return quick_string + character
	return (quick_string<charT, Traits, Alloc>(_Left) += _Right);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator==(
		const quick_string<charT, Traits, Alloc>& _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test for quick_string equality
	return (_Left.compare(_Right) == 0);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator==(
		const quick_string<charT, Traits, Alloc>& _Left,
		const typename quick_string<charT, Traits, Alloc>::string_type& _Right)
	{	// test for quick_string equality
	return (_Left.compare(_Right) == 0);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator==(
		const typename quick_string<charT, Traits, Alloc>::string_type& _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test for quick_string equality
	return (quick_string<charT, Traits, Alloc>(_Left.begin(), _Left.end()).compare(_Right) == 0);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator==(
		const charT * _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test for NTCS vs. quick_string equality
	return (_Right.compare(_Left) == 0);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator==(
		const quick_string<charT, Traits, Alloc>& _Left,
		const charT *_Right)
	{	// test for quick_string vs. NTCS equality
	return (_Left.compare(_Right) == 0);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator!=(
		const quick_string<charT, Traits, Alloc>& _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test for quick_string inequality
	return (!(_Left == _Right));
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator!=(
		const quick_string<charT, Traits, Alloc>& _Left,
		const typename quick_string<charT, Traits, Alloc>::string_type& _Right)
	{	// test for quick_string inequality
	return (!(_Left == _Right));
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator!=(
	const typename quick_string<charT, Traits, Alloc>::string_type& _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test for quick_string inequality
	return (!(_Left == _Right));
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator!=(
		const charT *_Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test for NTCS vs. quick_string inequality
	return (!(_Left == _Right));
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator!=(
		const quick_string<charT, Traits, Alloc>& _Left,
		const charT *_Right)
	{	// test for quick_string vs. NTCS inequality
	return (!(_Left == _Right));
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator<(
		const quick_string<charT, Traits, Alloc>& _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test if quick_string < quick_string
	return (_Left.compare(_Right) < 0);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator<(
		const quick_string<charT, Traits, Alloc>& _Left,
		const typename quick_string<charT, Traits, Alloc>::string_type& _Right)
	{	// test if quick_string < quick_string
	return (_Left.compare(_Right) < 0);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator<(
	const typename quick_string<charT, Traits, Alloc>::string_type& _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test if quick_string < quick_string
	return (quick_string<charT, Traits, Alloc>(_Left.begin(), _Left.end()).compare(_Right) < 0);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator<(
		const charT * _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test if NTCS < quick_string
	return (_Right.compare(_Left) > 0);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator<(
		const quick_string<charT, Traits, Alloc>& _Left,
		const charT *_Right)
	{	// test if quick_string < NTCS
	return (_Left.compare(_Right) < 0);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator>(
		const quick_string<charT, Traits, Alloc>& _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test if quick_string > quick_string
	return (_Right < _Left);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator>(
		const quick_string<charT, Traits, Alloc>& _Left,
		const typename quick_string<charT, Traits, Alloc>::string_type& _Right)
	{	// test if quick_string > quick_string
	return (_Right < _Left);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator>(
	const typename quick_string<charT, Traits, Alloc>::string_type& _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test if quick_string > quick_string
	return (_Right < _Left);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator>(
		const charT * _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test if NTCS > quick_string
	return (_Right < _Left);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator>(
		const quick_string<charT, Traits, Alloc>& _Left,
		const charT *_Right)
	{	// test if quick_string > NTCS
	return (_Right < _Left);
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator<=(
		const quick_string<charT, Traits, Alloc>& _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test if quick_string <= quick_string
	return (!(_Right < _Left));
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator<=(
		const quick_string<charT, Traits, Alloc>& _Left,
		const typename quick_string<charT, Traits, Alloc>::string_type& _Right)
	{	// test if quick_string <= quick_string
	return (!(_Right < _Left));
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator<=(
	const typename quick_string<charT, Traits, Alloc>::string_type& _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test if quick_string <= quick_string
	return (!(_Right < _Left));
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator<=(
		const charT * _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test if NTCS <= quick_string
	return (!(_Right < _Left));
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator<=(
		const quick_string<charT, Traits, Alloc>& _Left,
		const charT *_Right)
	{	// test if quick_string <= NTCS
	return (!(_Right < _Left));
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator>=(
		const quick_string<charT, Traits, Alloc>& _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test if quick_string >= quick_string
	return (!(_Left < _Right));
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator>=(
		const quick_string<charT, Traits, Alloc>& _Left,
		const typename quick_string<charT, Traits, Alloc>::string_type& _Right)
	{	// test if quick_string >= quick_string
	return (!(_Left < _Right));
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator>=(
		const typename quick_string<charT, Traits, Alloc>::string_type& _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test if quick_string >= quick_string
	return (!(_Left < _Right));
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator>=(
		const charT * _Left,
		const quick_string<charT, Traits, Alloc>& _Right)
	{	// test if NTCS >= quick_string
	return (!(_Left < _Right));
	}

template<class charT,
	class Traits,
	class Alloc> inline
	bool operator>=(
		const quick_string<charT, Traits, Alloc>& _Left,
		const charT *_Right)
	{	// test if quick_string >= NTCS
	return (!(_Left < _Right));
	}
}

#endif // LIB_QUICK_STRING
