// smart_ptr.h - v1.2.4
/*
 *  Copyright (c) 2010 Leigh Johnston.
 *
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions are
 *  met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *     * Neither the name of Leigh Johnston nor the names of any
 *       other contributors to this software may be used to endorse or
 *       promote products derived from this software without specific prior
 *       written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
 *  IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *  THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

/* WARNING: This class is not a substitute for the shared_ptr available on your platform
 * which you should be using instead.  It is here either for technical reasons or for when 
 * there is no standard library available.
  */

#ifndef LIB_SMART_PTR
#define LIB_SMART_PTR

namespace lib
{
	template <typename T>
	class smart_ptr
	{
	public:
		typedef T value_type;
		typedef T* pointer;
		typedef const T* const_pointer;
		typedef T& reference;
		typedef const T& const_reference;

	private:
		T* iPointer;
		typedef size_t counter_type;
		struct counter
		{
		public:
			counter() : iCounter(0) {}
			counter(pointer aPointer) : iCounter(0) 
			{
				try
				{
					iCounter = new counter_type;
				}
				catch(...)
				{
					delete aPointer;
					throw;
				}
				*iCounter = 0;
			}
			counter(const counter& aOther) : iCounter(aOther.iCounter) {}
			~counter() 
			{ 
				if (iCounter != 0 && *iCounter == 0) 
					delete iCounter; 
			}
			counter& operator=(const counter& aOther) 
			{ 
				iCounter = aOther.iCounter; 
				return *this;
			}
			counter_type operator++() 
			{ 
				if (iCounter != 0)
				{
					++*iCounter; 
					return *iCounter; 
				}
				else
					return 0;
			}
			counter_type operator--() 
			{ 
				if (iCounter != 0)
				{
					--*iCounter;
					return *iCounter;
				}
				else
					return 0;
			}
		private:
			counter_type* iCounter;
		} iReferenceCount;
	
	public:
		// construction
		smart_ptr() : iPointer(0) {}
		smart_ptr(pointer aPointer) : iPointer(aPointer), iReferenceCount(aPointer)
		{
			++iReferenceCount;
		}
		smart_ptr(const smart_ptr& aOther) : iPointer(aOther.iPointer), iReferenceCount(aOther.iReferenceCount)
		{
			++iReferenceCount;
		}
		~smart_ptr()
		{
			if (iPointer != 0)
				if (--iReferenceCount == 0)
					delete iPointer;
		}
		// element access
		reference operator*() { return *iPointer; }
		const_reference operator*() const { return *iPointer; }
		pointer operator->() { return iPointer; }
		const_pointer operator->() const { return iPointer; }
		pointer get() { return iPointer; }
		const_pointer get() const { return iPointer; }
		pointer release() 
		{ 
			pointer ret = iPointer;
			if (iPointer != 0)
			{
				iPointer = 0;
				--iReferenceCount;
			}
			return ret;
		}
		// modifiers
		void reset()
		{
			smart_ptr().swap(*this);
		}
		smart_ptr& operator=(pointer aPointer)
		{
			smart_ptr(aPointer).swap(*this);
			return *this;
		}
		smart_ptr& operator=(const smart_ptr& aOther)
		{
			smart_ptr(aOther).swap(*this);
			return *this;
		}
		void swap(smart_ptr& aOther)
		{
			std::swap(iPointer, aOther.iPointer);
			std::swap(iReferenceCount, aOther.iReferenceCount);
		}
		// comparison
		bool operator==(const smart_ptr& aOther) const
		{
			return get() == aOther.get();
		}
		bool operator!=(const smart_ptr& aOther) const
		{
			return get() != aOther.get();
		}
		bool operator<(const smart_ptr& aOther) const
		{
			return count() < aOther.count();
		}

	private:
		// implementation
		counter_type count() const
		{
			return iReferenceCount.count();
		}
	};

	template <typename T>
	inline smart_ptr<T> make_smart_ptr(T* aPointer)
	{
		return smart_ptr<T>(aPointer);
	}
}

#endif // LIB_SMART_PTR
